# stdlib
from typing import Optional

# 3rd party
from docutils import nodes
from domdf_python_tools.paths import PathPlus
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.locale import admonitionlabels
from sphinx.util.docutils import SphinxRole
from sphinx.writers.latex import LaTeXTranslator
from sphinx_toolbox import latex


def visit_seealso(translator: LaTeXTranslator, node: addnodes.seealso) -> None:
	"""
	Visit an :class:`addnodes.seealso`` node.

	:param translator:
	:param node:
	"""

	if len(node) > 1:
		LaTeXTranslator.visit_seealso(translator, node)
	else:
		translator.body.append('\n\n\\sphinxstrong{%s:} ' % admonitionlabels["seealso"])


def replace_latex(app: Sphinx, exception: Optional[Exception] = None):
	if exception:
		return

	if app.builder.name.lower() != "latex":
		return

	output_file = PathPlus(app.builder.outdir) / f"{app.builder.titles[0][1]}.tex"

	output_content = output_file.read_text()

	output_content = output_content.replace('≡', r"$\equiv$")
	output_content = output_content.replace(
			r"@\spxentry{verify}\spxextra{SlumberURL attribute}}\needspace{5\baselineskip}",
			r"@\spxentry{verify}\spxextra{SlumberURL attribute}}",
			)

	output_file.write_clean(output_content)


class RequestsXRef(SphinxRole):

	def run(self):
		return [
				nodes.reference(
						"`Requests <https://requests.readthedocs.io>`_",
						"Requests",
						refuri="https://requests.readthedocs.io"
						)
				], []


def setup(app: Sphinx):
	app.connect("build-finished", replace_latex)
	app.connect("build-finished", latex.replace_unknown_unicode)
	app.add_node(addnodes.seealso, latex=(visit_seealso, LaTeXTranslator.depart_seealso), override=True)
	app.add_role("requests", RequestsXRef())



#!/usr/bin/env python3

# This file is managed by 'repo_helper'. Don't edit it directly.

# stdlib
import os
import re
import sys

# 3rd party
from sphinx_pyproject import SphinxConfig

sys.path.append('.')

config = SphinxConfig(globalns=globals())
project = config["project"]
author = config["author"]
documentation_summary = config.description

github_url = "https://github.com/{github_username}/{github_repository}".format_map(config)

rst_prolog = f""".. |pkgname| replace:: apeye
.. |pkgname2| replace:: ``apeye``
.. |browse_github| replace:: `Browse the GitHub Repository <{github_url}>`__
"""

slug = re.sub(r'\W+', '-', project.lower())
release = version = config.version

sphinx_builder = os.environ.get("SPHINX_BUILDER", "html").lower()
todo_include_todos = int(os.environ.get("SHOW_TODOS", 0)) and sphinx_builder != "latex"

intersphinx_mapping = {
		"python": ("https://docs.python.org/3/", None),
		"sphinx": ("https://www.sphinx-doc.org/en/stable/", None),
		}

html_theme_options = {
		"light_css_variables": {
				"toc-title-font-size": "12pt",
				"toc-font-size": "12pt",
				"admonition-font-size": "12pt",
				},
		"dark_css_variables": {
				"toc-title-font-size": "12pt",
				"toc-font-size": "12pt",
				"admonition-font-size": "12pt",
				},
		}

html_context = {}
htmlhelp_basename = slug

latex_documents = [("index", f'{slug}.tex', project, author, "manual")]
man_pages = [("index", slug, project, [author], 1)]
texinfo_documents = [("index", slug, project, author, slug, project, "Miscellaneous")]

toctree_plus_types = set(config["toctree_plus_types"])

autodoc_default_options = {
		"members": None,  # Include all members (methods).
		"special-members": None,
		"autosummary": None,
		"show-inheritance": None,
		"exclude-members": ','.join(config["autodoc_exclude_members"]),
		}

latex_elements = {
		"printindex": "\\begin{flushleft}\n\\printindex\n\\end{flushleft}",
		"tableofcontents": "\\pdfbookmark[0]{\\contentsname}{toc}\\sphinxtableofcontents",
		}


def setup(app):
	# 3rd party
	from sphinx_toolbox.latex import better_header_layout

	app.connect("config-inited", lambda app, config: better_header_layout(config))


nitpicky = True
needspace_amount = r"5\baselineskip"
ignore_missing_xrefs = [
		"^requests.structures.CaseInsensitiveDict$",
		"^cachecontrol.adapter.CacheControlAdapter",
		]



# stdlib
import inspect
from typing import Any, Callable, Dict, List

# 3rd party
from sphinx.application import Sphinx
from sphinx_toolbox.more_autodoc import generic_bases
from sphinx_toolbox.more_autodoc.overloads import FunctionDocumenter
from sphinx_toolbox.more_autosummary import PatchedAutoSummClassDocumenter
from sphinx_toolbox.utils import SphinxExtMetadata


def alias_option(arg: Any) -> List[str]:
	"""
	Used to convert the :members: option to auto directives.
	"""

	if arg is None:
		return []
	else:
		return [x.strip() for x in arg.split(',') if x.strip()]


class AliasedClassDocumenter(generic_bases.GenericBasesClassDocumenter):

	option_spec: Dict[str, Callable] = {
			**PatchedAutoSummClassDocumenter.option_spec,
			"autoclass-alias": alias_option,
			}

	def add_directive_header(self, sig: str) -> None:
		"""
		Add the directive's header, and the inheritance information if the ``:show-inheritance:`` flag set.

		:param sig: The NamedTuple's signature.
		"""

		if "autoclass-alias" not in self.options:
			return super().add_directive_header(sig)

		if self.doc_as_attr:
			self.directivetype = "attribute"

		domain = getattr(self, "domain", "py")
		directive = getattr(self, "directivetype", self.objtype)
		name = self.format_name()
		sourcename = self.get_sourcename()

		if '\n' in sig:
			raise NotImplementedError(sig)

		# one signature per line, indented by column
		prefix = f'.. {domain}:{directive}:: '
		for i, sig_line in enumerate(sig.split('\n')):
			self.add_line(f'{prefix}{name}{sig_line}', sourcename)
			if i == 0:
				prefix = ' ' * len(prefix)

		# for i, alias in enumerate(self.options["autoclass-alias"]):
		# 	self.add_line(f'{prefix}{alias}{sig}', sourcename)
		# 	if i == 0:
		# 		prefix = ' ' * len(prefix)
		#
		# if self.objpath:
		# 	self.add_line(f'{prefix}{name}{sig}', sourcename)

		if self.options.noindex:
			self.add_line("   :noindex:", sourcename)

		if self.objpath:
			# Be explicit about the module, this is necessary since .. class::
			# etc. don't support a prepended module name
			self.add_line(f"   :module: {self.modname}", sourcename)

		self.add_line(f"   :canonical: {self.options['autoclass-alias'][0]}", sourcename)

		if self.analyzer and '.'.join(self.objpath) in self.analyzer.finals:
			self.add_line("   :final:", sourcename)

		# add inheritance info, if wanted
		if not self.doc_as_attr and self.options.show_inheritance:
			generic_bases._add_generic_bases(self)


class AliasedFunctionDocumenter(FunctionDocumenter):

	option_spec: Dict[str, Callable] = {
			**FunctionDocumenter.option_spec,
			"autofunction-alias": alias_option,
			}

	def add_directive_header(self, sig: str) -> None:
		"""
		Add the directive's header.

		:param sig:
		"""

		if "autofunction-alias" not in self.options:
			return super().add_directive_header(sig)

		domain = getattr(self, "domain", "py")
		directive = getattr(self, "directivetype", self.objtype)
		name = self.format_name()
		sourcename = self.get_sourcename()

		if '\n' in sig:
			raise NotImplementedError(sig)

		# one signature per line, indented by column
		prefix = f".. {domain}:{directive}:: "
		for i, sig_line in enumerate(sig.split('\n')):
			self.add_line(f'{prefix}{name}{sig_line}', sourcename)
			if i == 0:
				prefix = ' ' * len(prefix)

		# for i, alias in enumerate(self.options["autofunction-alias"]):
		# 	self.add_line(f'{prefix}{alias}{sig}', sourcename)
		# 	if i == 0:
		# 		prefix = ' ' * len(prefix)
		#
		# if self.objpath:
		# 	self.add_line(f'{prefix}{name}{sig}', sourcename)

		if self.options.noindex:
			self.add_line("   :noindex:", sourcename)

		if self.objpath:
			# Be explicit about the module, this is necessary since .. class::
			# etc. don't support a prepended module name
			self.add_line(f"   :module: {self.modname}", sourcename)

		self.add_line(f"   :canonical: {self.options['autofunction-alias'][0]}", sourcename)

		if inspect.iscoroutinefunction(self.object):
			self.add_line("   :async:", sourcename)

		if self.env.config.overloads_location == "top":
			for line in self.create_body_overloads():
				self.add_line(f"{self.content_indent}{line}", self.get_sourcename())


def setup(app: Sphinx) -> SphinxExtMetadata:
	app.add_autodocumenter(AliasedClassDocumenter, override=True)
	app.add_autodocumenter(AliasedFunctionDocumenter, override=True)
	return {"parallel_read_safe": True}



# stdlib
import json
import sys

# 3rd party
import pytest
from coincidence import AdvancedFileRegressionFixture
from coincidence.params import param
from domdf_python_tools.compat import PYPY36

# this package
from apeye.email_validator import EmailSyntaxError, ValidatedEmail
from apeye.email_validator import main as validator_main
from apeye.email_validator import validate_email


@pytest.mark.parametrize(
		"email_input,output",
		[
				(
						"Abc@example.com",
						ValidatedEmail(
								local_part="Abc",
								ascii_local_part="Abc",
								smtputf8=False,
								ascii_domain="example.com",
								domain="example.com",
								email="Abc@example.com",
								original_email="Abc@example.com",
								ascii_email="Abc@example.com",
								),
						),
				(
						b"Abc@example.com",
						ValidatedEmail(
								local_part="Abc",
								ascii_local_part="Abc",
								smtputf8=False,
								ascii_domain="example.com",
								domain="example.com",
								email="Abc@example.com",
								original_email="Abc@example.com",
								ascii_email="Abc@example.com",
								),
						),
				(
						"Abc.123@example.com",
						ValidatedEmail(
								local_part="Abc.123",
								ascii_local_part="Abc.123",
								smtputf8=False,
								ascii_domain="example.com",
								domain="example.com",
								email="Abc.123@example.com",
								original_email="Abc.123@example.com",
								ascii_email="Abc.123@example.com",
								),
						),
				(
						b"Abc.123@example.com",
						ValidatedEmail(
								local_part="Abc.123",
								ascii_local_part="Abc.123",
								smtputf8=False,
								ascii_domain="example.com",
								domain="example.com",
								email="Abc.123@example.com",
								original_email="Abc.123@example.com",
								ascii_email="Abc.123@example.com",
								),
						),
				(
						"user+mailbox/department=shipping@example.com",
						ValidatedEmail(
								local_part="user+mailbox/department=shipping",
								ascii_local_part="user+mailbox/department=shipping",
								smtputf8=False,
								ascii_domain="example.com",
								domain="example.com",
								email="user+mailbox/department=shipping@example.com",
								original_email="user+mailbox/department=shipping@example.com",
								ascii_email="user+mailbox/department=shipping@example.com",
								),
						),
				(
						b"user+mailbox/department=shipping@example.com",
						ValidatedEmail(
								local_part="user+mailbox/department=shipping",
								ascii_local_part="user+mailbox/department=shipping",
								smtputf8=False,
								ascii_domain="example.com",
								domain="example.com",
								email="user+mailbox/department=shipping@example.com",
								original_email="user+mailbox/department=shipping@example.com",
								ascii_email="user+mailbox/department=shipping@example.com",
								),
						),
				(
						"!#$%&'*+-/=?^_`.{|}~@example.com",
						ValidatedEmail(
								local_part="!#$%&'*+-/=?^_`.{|}~",
								ascii_local_part="!#$%&'*+-/=?^_`.{|}~",
								smtputf8=False,
								ascii_domain="example.com",
								domain="example.com",
								email="!#$%&'*+-/=?^_`.{|}~@example.com",
								original_email="!#$%&'*+-/=?^_`.{|}~@example.com",
								ascii_email="!#$%&'*+-/=?^_`.{|}~@example.com",
								),
						),
				(
						"伊昭傑@郵件.商務",
						ValidatedEmail(
								local_part="伊昭傑",
								smtputf8=True,
								ascii_domain="xn--5nqv22n.xn--lhr59c",
								domain="郵件.商務",
								email="伊昭傑@郵件.商務",
								original_email="伊昭傑@郵件.商務",
								),
						),
				(
						"राम@मोहन.ईन्फो",
						ValidatedEmail(
								local_part="राम",
								smtputf8=True,
								ascii_domain="xn--l2bl7a9d.xn--o1b8dj2ki",
								domain="मोहन.ईन्फो",
								email="राम@मोहन.ईन्फो",
								original_email="राम@मोहन.ईन्फो",
								),
						),
				(
						"юзер@екзампл.ком",
						ValidatedEmail(
								local_part="юзер",
								smtputf8=True,
								ascii_domain="xn--80ajglhfv.xn--j1aef",
								domain="екзампл.ком",
								email="юзер@екзампл.ком",
								original_email="юзер@екзампл.ком",
								),
						),
				(
						"θσερ@εχαμπλε.ψομ",
						ValidatedEmail(
								local_part="θσερ",
								smtputf8=True,
								ascii_domain="xn--mxahbxey0c.xn--xxaf0a",
								domain="εχαμπλε.ψομ",
								email="θσερ@εχαμπλε.ψομ",
								original_email="θσερ@εχαμπλε.ψομ",
								),
						),
				(
						"葉士豪@臺網中心.tw",
						ValidatedEmail(
								local_part="葉士豪",
								smtputf8=True,
								ascii_domain="xn--fiqq24b10vi0d.tw",
								domain="臺網中心.tw",
								email="葉士豪@臺網中心.tw",
								original_email="葉士豪@臺網中心.tw",
								),
						),
				(
						"jeff@臺網中心.tw",
						ValidatedEmail(
								local_part="jeff",
								ascii_local_part="jeff",
								smtputf8=False,
								ascii_domain="xn--fiqq24b10vi0d.tw",
								domain="臺網中心.tw",
								email="jeff@臺網中心.tw",
								original_email="jeff@臺網中心.tw",
								ascii_email="jeff@xn--fiqq24b10vi0d.tw",
								),
						),
				(
						"葉士豪@臺網中心.台灣",
						ValidatedEmail(
								local_part="葉士豪",
								smtputf8=True,
								ascii_domain="xn--fiqq24b10vi0d.xn--kpry57d",
								domain="臺網中心.台灣",
								email="葉士豪@臺網中心.台灣",
								original_email="葉士豪@臺網中心.台灣",
								),
						),
				(
						"jeff葉@臺網中心.tw",
						ValidatedEmail(
								local_part="jeff葉",
								smtputf8=True,
								ascii_domain="xn--fiqq24b10vi0d.tw",
								domain="臺網中心.tw",
								email="jeff葉@臺網中心.tw",
								original_email="jeff葉@臺網中心.tw",
								),
						),
				(
						"ñoñó@example.com",
						ValidatedEmail(
								local_part="ñoñó",
								smtputf8=True,
								ascii_domain="example.com",
								domain="example.com",
								email="ñoñó@example.com",
								original_email="ñoñó@example.com",
								),
						),
				(
						"我買@example.com",
						ValidatedEmail(
								local_part="我買",
								smtputf8=True,
								ascii_domain="example.com",
								domain="example.com",
								email="我買@example.com",
								original_email="我買@example.com",
								),
						),
				(
						"甲斐黒川日本@example.com",
						ValidatedEmail(
								local_part="甲斐黒川日本",
								smtputf8=True,
								ascii_domain="example.com",
								domain="example.com",
								email="甲斐黒川日本@example.com",
								original_email="甲斐黒川日本@example.com",
								),
						),
				param(
						"чебурашкаящик-с-апельсинами.рф@example.com",
						ValidatedEmail(
								local_part="чебурашкаящик-с-апельсинами.рф",
								smtputf8=True,
								ascii_domain="example.com",
								domain="example.com",
								email="чебурашкаящик-с-апельсинами.рф@example.com",
								original_email="чебурашкаящик-с-апельсинами.рф@example.com",
								),
						marks=pytest.mark.skipif(
								sys.platform == "win32" and PYPY36,
								reason="Fails due to unicode issue with filename",
								),
						id="unicode_1",
						),
				(
						"उदाहरण.परीक्ष@domain.with.idn.tld",
						ValidatedEmail(
								local_part="उदाहरण.परीक्ष",
								smtputf8=True,
								ascii_domain="domain.with.idn.tld",
								domain="domain.with.idn.tld",
								email="उदाहरण.परीक्ष@domain.with.idn.tld",
								original_email="उदाहरण.परीक्ष@domain.with.idn.tld",
								),
						),
				(
						"ιωάννης@εεττ.gr",
						ValidatedEmail(
								local_part="ιωάννης",
								smtputf8=True,
								ascii_domain="xn--qxaa9ba.gr",
								domain="εεττ.gr",
								email="ιωάννης@εεττ.gr",
								original_email="ιωάννης@εεττ.gr",
								),
						),
				],
		)
def test_email_valid(
		email_input: str,
		output: ValidatedEmail,
		advanced_file_regression: AdvancedFileRegressionFixture,
		):
	# print(f'({email_input!r}, {validate_email(email_input)!r}),')
	assert validate_email(email_input) == output

	data = output.as_dict()
	data["__repr__"] = repr(output)
	data["__str__"] = str(output)
	advanced_file_regression.check(json.dumps(data), extension=".json")


@pytest.mark.parametrize(
		"email_input,error_msg",
		[
				("my@.leadingdot.com", "An email address cannot have a period immediately after the @-sign."),
				("my@．．leadingfwdot.com", "An email address cannot have a period immediately after the @-sign."),
				("my@..twodots.com", "An email address cannot have a period immediately after the @-sign."),
				("my@twodots..com", "An email address cannot have two periods in a row."),
				(
						"my@baddash.-.com",
						{
								"The domain name baddash.-.com contains invalid characters (Label must not start or end with a hyphen).",
								"The domain name baddash.-.com contains invalid characters (The label b'-' is not a valid A-label).",
								}
						),
				(
						"my@baddash.-a.com",
						{
								"The domain name baddash.-a.com contains invalid characters (Label must not start or end with a hyphen).",
								"The domain name baddash.-a.com contains invalid characters (The label b'-a' is not a valid A-label).",
								}
						),
				(
						"my@baddash.b-.com",
						{
								"The domain name baddash.b-.com contains invalid characters (Label must not start or end with a hyphen).",
								"The domain name baddash.b-.com contains invalid characters (The label b'b-' is not a valid A-label)."
								}
						),
				(
						'my@example.com\n',
						{
								'The domain name example.com\n contains invalid characters (Codepoint U+000A at position 4 of \'com\\n\' not allowed).',
								"The domain name example.com\n contains invalid characters (Codepoint U+000A not allowed at position 12 in 'example.com\\n').",
								}
						),
				(
						'my@example\n.com',
						{
								"The domain name example\n.com contains invalid characters (Codepoint U+000A at position 8 of 'example\\n' not allowed).",
								"The domain name example\n.com contains invalid characters (Codepoint U+000A not allowed at position 8 in 'example\\n.com').",
								}
						),
				(".leadingdot@domain.com", "The email address contains invalid characters before the @-sign: .."),
				("..twodots@domain.com", "The email address contains invalid characters before the @-sign: .."),
				(
						"twodots..here@domain.com",
						"The email address contains invalid characters before the @-sign: .."
						),
				(
						"me@⒈wouldbeinvalid.com",
						"The domain name ⒈wouldbeinvalid.com contains invalid characters (Codepoint U+2488 not allowed "
						"at position 1 in '⒈wouldbeinvalid.com')."
						),
				("@example.com", "There must be something before the @-sign."),
				('\nmy@example.com', 'The email address contains invalid characters before the @-sign: \n.'),
				('m\ny@example.com', 'The email address contains invalid characters before the @-sign: \n.'),
				('my\n@example.com', 'The email address contains invalid characters before the @-sign: \n.'),
				(
						"11111111112222222222333333333344444444445555555555666666666677777@example.com",
						"The email address is too long before the @-sign (1 character too many)."
						),
				(
						"111111111122222222223333333333444444444455555555556666666666777777@example.com",
						"The email address is too long before the @-sign (2 characters too many)."
						),
				(
						"me@1111111111222222222233333333334444444444555555555.6666666666777777777788888888889999999999000000000.1111111111222222222233333333334444444444555555555.6666666666777777777788888888889999999999000000000.111111111122222222223333333333444444444455555555556.com",
						"The email address is too long after the @-sign."
						),
				(
						"my.long.address@1111111111222222222233333333334444444444555555555.6666666666777777777788888888889999999999000000000.1111111111222222222233333333334444444444555555555.6666666666777777777788888888889999999999000000000.11111111112222222222333333333344444.info",
						"The email address is too long (2 characters too many)."
						),
				(
						"my.long.address@λ111111111222222222233333333334444444444555555555.6666666666777777777788888888889999999999000000000.1111111111222222222233333333334444444444555555555.6666666666777777777788888888889999999999000000000.11111111112222222222333333.info",
						"The email address is too long (when converted to IDNA ASCII)."
						),
				(
						"my.long.address@λ111111111222222222233333333334444444444555555555.6666666666777777777788888888889999999999000000000.1111111111222222222233333333334444444444555555555.6666666666777777777788888888889999999999000000000.1111111111222222222233333333334444.info",
						"The email address is too long (at least 1 character too many)."
						),
				(
						"my.λong.address@1111111111222222222233333333334444444444555555555.6666666666777777777788888888889999999999000000000.1111111111222222222233333333334444444444555555555.6666666666777777777788888888889999999999000000000.111111111122222222223333333333444.info",
						"The email address is too long (when encoded in bytes)."
						),
				(
						"my.λong.address@1111111111222222222233333333334444444444555555555.6666666666777777777788888888889999999999000000000.1111111111222222222233333333334444444444555555555.6666666666777777777788888888889999999999000000000.1111111111222222222233333333334444.info",
						"The email address is too long (at least 1 character too many)."
						),
				("local_part_only@", "There must be something after the @-sign."),
				("dom@example.com.", "An email address cannot end with a period."),
				],
		)
def test_email_invalid(email_input, error_msg):
	with pytest.raises(EmailSyntaxError) as exc_info:
		validate_email(email_input)
	# print(f'({email_input!r}, {str(exc_info.value)!r}),')

	if isinstance(error_msg, set):
		assert str(exc_info.value) in error_msg
	else:
		assert str(exc_info.value) == error_msg


def test_dict_accessor():
	input_email = "testaddr@example.com"
	valid_email = validate_email(input_email)
	assert isinstance(valid_email.as_dict(), dict)
	assert valid_email.as_dict()["original_email"] == input_email


def test_main_single_good_input(monkeypatch, capsys):
	# stdlib
	import json
	test_email = "test@example.com"
	monkeypatch.setattr("sys.argv", ["email_validator", test_email])
	validator_main()
	stdout, _ = capsys.readouterr()
	output = json.loads(str(stdout))
	assert isinstance(output, dict)
	assert validate_email(test_email).original_email == output["original_email"]


def test_main_single_bad_input(monkeypatch, capsys):
	bad_email = "test@..com"
	monkeypatch.setattr("sys.argv", ["email_validator", bad_email])
	validator_main()
	stdout, _ = capsys.readouterr()
	assert stdout == 'An email address cannot have a period immediately after the @-sign.\n'


def test_main_multi_input(monkeypatch, capsys):
	# stdlib
	import io
	test_cases = ["test@example.com", "test2@example.com", "test@.com", "test3@.com"]
	test_input = io.StringIO('\n'.join(test_cases))
	monkeypatch.setattr("sys.stdin", test_input)
	monkeypatch.setattr("sys.argv", ["email_validator"])
	validator_main()
	stdout, _ = capsys.readouterr()
	assert test_cases[0] not in stdout
	assert test_cases[1] not in stdout
	assert test_cases[2] in stdout
	assert test_cases[3] in stdout



# stdlib
import json
import logging
import re
import sys
import time
from datetime import datetime

# 3rd party
import pytest
import requests
from pytest_httpserver import HTTPServer
from werkzeug import Request, Response

# this package
from apeye.rate_limiter import HTTPCache, rate_limit

if sys.version_info < (3, 7):
	# 3rd party
	from backports.datetime_fromisoformat import datetime_fromisoformat

else:
	datetime_fromisoformat = datetime.fromisoformat

logging.basicConfig()


@rate_limit(3)
def rate_limited_function():
	print("Inside function")
	return 42


def test_rate_limit(capsys, caplog):
	assert rate_limited_function() == 42
	assert rate_limited_function() == 42
	assert rate_limited_function() == 42

	assert capsys.readouterr().out.splitlines() == ["Inside function"] * 3
	caplog.set_level(logging.DEBUG)

	assert rate_limited_function() == 42
	assert rate_limited_function() == 42
	time.sleep(2)
	assert rate_limited_function() == 42

	assert capsys.readouterr().out.splitlines() == ["Inside function"] * 3

	last_ran_re = re.compile(r"rate_limited_function: Last ran (\d+(\.\d+)?|\d+e-\d+) seconds ago\.")

	print(caplog.record_tuples)

	assert last_ran_re.match(caplog.record_tuples[0][2])
	assert re.match(r"rate_limited_function: Waiting (\d+(\.\d+)?|\d+e-\d+) seconds\.", caplog.record_tuples[1][2])
	assert last_ran_re.match(caplog.record_tuples[2][2])
	assert last_ran_re.match(caplog.record_tuples[3][2])


@pytest.fixture(scope="session")
def testing_http_cache():
	cache = HTTPCache("testing_apeye_http")
	assert cache.clear()
	yield cache

	cache.clear()


@pytest.fixture()
def timeserver(httpserver: HTTPServer):

	def time_handler(request: Request):
		time.sleep(1)

		now = datetime.utcnow()
		headers = {
				"Cache-Control": "max-age=0, private, must-revalidate",
				"Date": now.strftime("%a, %d %B %Y %H:%M:%S GMT"),
				}
		response_json = {
				"abbreviation": "GMT",
				"client_ip": "127.0.0.1",
				"datetime": now.strftime("%Y-%m-%dT%H:%M:%S%z"),
				"day_of_week": now.strftime("%w"),
				"day_of_year": now.strftime("%j"),
				}

		return Response(json.dumps(response_json, indent=4), 200, headers, None, None)

	httpserver.expect_request("/time", method="GET").respond_with_handler(time_handler)

	return httpserver


def test_cache_canary(timeserver: HTTPServer):
	# Proves that worldtimeapi.org returns a different time for two sequential requests.
	session = requests.session()

	# target_url = "http://worldtimeapi.org/api/ip"
	target_url = timeserver.url_for("/time")

	response = session.get(target_url)
	assert response.status_code == 200
	original_time = datetime_fromisoformat(response.json()["datetime"])

	response = session.get(target_url)
	assert response.status_code == 200
	current_time = datetime_fromisoformat(response.json()["datetime"])

	assert current_time > original_time


@pytest.mark.parametrize("run_number", [1, 2])
def test_http_cache(testing_http_cache, capsys, run_number: int, timeserver: HTTPServer):
	session = testing_http_cache.session

	# target_url = "http://worldtimeapi.org/api/ip"
	target_url = timeserver.url_for("/time")

	response = session.get(target_url)
	assert response.status_code == 200
	assert not response.from_cache
	original_time = datetime_fromisoformat(response.json()["datetime"])

	response = session.get(target_url)
	assert response.status_code == 200
	assert response.from_cache
	current_time = datetime_fromisoformat(response.json()["datetime"])

	# If the times have changed the cache has failed.
	assert current_time == original_time

	assert testing_http_cache.cache_dir.is_dir()
	assert testing_http_cache.clear()
	assert not testing_http_cache.cache_dir.is_dir()

	# make a new request
	response = session.get(target_url)
	assert response.status_code == 200
	assert not response.from_cache
	current_time = datetime_fromisoformat(response.json()["datetime"])

	assert current_time > original_time

	assert testing_http_cache.clear()



# stdlib
import doctest
import inspect
import shutil
from pathlib import PurePosixPath
from textwrap import indent

# 3rd party
import pytest
from domdf_python_tools.utils import redirect_output

# this package
import apeye.url

VERBOSE = 1

ret = 0


def relative_to(self, *other):
	# From Python 3.8, pathlib.py
	if not other:
		raise TypeError("need at least one argument")
	parts = self._parts
	drv = self._drv
	root = self._root
	if root:
		abs_parts = [drv, root] + parts[1:]
	else:
		abs_parts = parts
	to_drv, to_root, to_parts = self._parse_args(other)
	if to_root:
		to_abs_parts = [to_drv, to_root] + to_parts[1:]
	else:
		to_abs_parts = to_parts
	n = len(to_abs_parts)
	cf = self._flavour.casefold_parts
	if (root or drv) if n == 0 else cf(abs_parts[:n]) != cf(to_abs_parts):
		formatted = self._format_parsed_parts(to_drv, to_root, to_parts)
		raise ValueError(f"{str(self)!r} does not start with {str(formatted)!r}")
	return self._from_parsed_parts('', root if n == 1 else '', abs_parts[n:])


@pytest.mark.parametrize("module", [apeye.url])
def test_docstrings(module, monkeypatch):

	# Always use pre-Python 3.9 implementation
	monkeypatch.setattr(PurePosixPath, "relative_to", relative_to)

	# Check that we were actually given a module.
	if inspect.ismodule(module):
		print(f"Running doctest in {module!r}".center(shutil.get_terminal_size().columns, '='))
	else:
		raise TypeError(f"testmod: module required; {module!r}")

	with redirect_output(combine=True) as (stdout, stderr):

		# Find, parse, and run all tests in the given module.
		finder = doctest.DocTestFinder()
		runner = doctest.DocTestRunner(verbose=VERBOSE >= 2)

		for test in finder.find(module, module.__name__):
			runner.run(test)

		runner.summarize(verbose=bool(VERBOSE))

	# results = doctest.TestResults(runner.failures, runner.tries)
	print(indent(stdout.getvalue(), "  "))

	if runner.failures:
		pytest.fail(msg=f"{runner.failures} tests failed")



# 3rd party
import pytest

# this package
from apeye.cache import Cache


@pytest.fixture(scope="session")
def testing_cache():
	cache = Cache("testing_apeye")
	assert cache.clear()
	yield cache
	assert cache.clear()


@pytest.mark.parametrize("run_number", [1, 2])
def test_cache(testing_cache, capsys, run_number):

	@testing_cache
	def cached_function(arg1: int, arg2: float, arg3: str):
		print("Running")
		return (arg1**int(arg2)) * arg3

	assert not (testing_cache.cache_dir / "cached_function.json").is_file()

	assert cached_function(2, 5.5, '☃') == '☃' * 32
	assert (testing_cache.cache_dir / "cached_function.json").is_file()

	for i in range(10):
		assert cached_function(2, 5.5, '☃') == '☃' * 32

	assert cached_function(2, 5.8, '☃') == '☃' * 32

	captured = capsys.readouterr()
	assert captured.out == "Running\nRunning\n"

	assert testing_cache.clear(cached_function)
	assert not (testing_cache.cache_dir / "cached_function.json").is_file()
	assert testing_cache.cache_dir.is_dir()

	assert cached_function(2, 5.5, '☃') == '☃' * 32
	capsys.readouterr()  # prevents the above call polluting stdout

	old_id = id(cached_function)

	@testing_cache  # type: ignore[no-redef]
	def cached_function(arg1: int, arg2: float, arg3: str):
		print("Running 2nd function")
		return (arg1**int(arg2)) * arg3

	assert id(cached_function) != old_id

	assert cached_function(2, 5.5, '☃') == '☃' * 32
	assert (testing_cache.cache_dir / "cached_function.json").is_file()
	assert testing_cache.clear(cached_function)

	captured = capsys.readouterr()
	# if the cache wasn't working this would be "Running 2nd function\n"
	assert captured.out == ''



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0
#
# From https://github.com/ionrock/cachecontrol

# 3rd party
import pytest
from cachecontrol.cache import DictCache
from requests import Session

# this package
from apeye.rate_limiter import RateLimitAdapter


class NullSerializer:

	def dumps(self, request, response, body=None):
		return response

	def loads(self, request, data, body_file=None):
		if data and getattr(data, "chunked", False):
			data.chunked = False
		return data


class TestMaxAge:

	@pytest.fixture()
	def sess(self, url):
		self.url = url
		self.cache = DictCache()
		sess = Session()
		adapter = RateLimitAdapter(
				self.cache,
				serializer=NullSerializer()  # type: ignore[arg-type]  # NullSerializer is not the right type
				)
		sess.mount("http://", adapter)
		return sess

	def test_client_max_age_0(self, sess):
		"""
		Making sure when the client uses max-age=0 we don't get a
		cached copy even though we're still fresh.
		"""
		print("first request")
		r = sess.get(self.url)
		assert self.cache.get(self.url) == r.raw

		print("second request")
		r = sess.get(self.url, headers={"Cache-Control": "max-age=0"})

		# don't remove from the cache
		assert self.cache.get(self.url)
		assert not r.from_cache

	def test_client_max_age_3600(self, sess):
		"""
		Verify we get a cached value when the client has a
		reasonable max-age value.
		"""
		r = sess.get(self.url)
		assert self.cache.get(self.url) == r.raw

		# request that we don't want a new one unless
		r = sess.get(self.url, headers={"Cache-Control": "max-age=3600"})
		assert r.from_cache is True

		# now lets grab one that forces a new request b/c the cache
		# has expired. To do that we'll inject a new time value.
		resp = self.cache.get(self.url)
		assert resp is not None
		resp.headers[  # type: ignore[attr-defined]  # something is wrong in CacheControl's new type hints
				"date"] = "Tue, 15 Nov 1994 08:12:31 GMT"
		r = sess.get(self.url)
		assert not r.from_cache



# stdlib
import json
import os
import pathlib
import sys
from abc import ABC, abstractmethod
from ipaddress import IPv4Address
from textwrap import dedent
from typing import Type
from urllib.parse import parse_qs

# 3rd party
import pytest
import pytest_httpserver.pytest_plugin
import requests
import werkzeug
from coincidence import count, not_pypy
from pytest_httpserver.httpserver import QueryMatcher

# this package
from apeye.requests_url import RequestsURL, TrailingRequestsURL
from apeye.slumber_url import SlumberURL
from apeye.url import URL, Domain, URLPath


class MyPathLike(os.PathLike):

	def __fspath__(self):
		return "/python"


class TestURLPath:

	@pytest.mark.parametrize(
			"value, expected",
			[
					("/watch?v=NG21KWZSiok", '/'),
					("watch?v=NG21KWZSiok", ''),
					('', ''),
					("/programmes/b006qtlx/episodes/player", '/'),
					("/news", '/'),
					("news", ''),
					]
			)
	def test_drive_root_anchor(self, value: str, expected: str):
		# These were originally NotImplemented, but they weren't doing any harm and might be useful
		assert URLPath(value).drive == ''
		assert URLPath(value).root == expected
		assert URLPath(value).anchor == expected

	@pytest.mark.parametrize(
			"value, absolute",
			[
					(URLPath(), False),
					(URLPath(''), False),
					(URLPath('/'), True),
					(URLPath("/news"), True),
					(URLPath("news"), False),
					(URLPath("/programmes/b006qtlx"), True),
					(URLPath("programmes/b006qtlx"), False),
					(URLPath("/programmes/b006qtlx/episodes/player"), True),
					(URLPath("/watch?v=NG21KWZSiok"), True),
					(URLPath("watch?v=NG21KWZSiok"), False),
					]
			)
	def test_is_absolute(self, value: str, absolute: bool):
		# Was originally NotImplemented, but it might be useful
		assert URLPath(value).is_absolute() is absolute

	@pytest.mark.parametrize(
			"value", [
					"/watch?v=NG21KWZSiok",
					"watch?v=NG21KWZSiok",
					'',
					"/programmes/b006qtlx/episodes/player",
					]
			)
	def test_str(self, value):
		assert str(URLPath(value)) == value

	@pytest.mark.parametrize(
			"value, expects",
			[
					("/watch?v=NG21KWZSiok", "URLPath('/watch?v=NG21KWZSiok')"),
					("watch?v=NG21KWZSiok", "URLPath('watch?v=NG21KWZSiok')"),
					('', "URLPath('')"),
					("/programmes/b006qtlx/episodes/player", "URLPath('/programmes/b006qtlx/episodes/player')"),
					]
			)
	def test_repr(self, value, expects):
		assert repr(URLPath(value)) == expects

	@pytest.mark.parametrize("method", [URLPath().as_uri])
	def test_notimplemented(self, method):
		with pytest.raises(NotImplementedError):
			method()

	@pytest.mark.parametrize(
			"value, expects",
			[
					(URLPath() / "news", URLPath("news")),
					(URLPath('/') / "news", URLPath("/news")),
					(URLPath("/programmes") / "b006qtlx", URLPath("/programmes/b006qtlx")),
					("/programmes" / URLPath("b006qtlx"), URLPath("/programmes/b006qtlx")),
					]
			)
	def test_division(self, value, expects):
		assert value == expects
		assert isinstance(value, URLPath)

	@pytest.mark.parametrize(
			"value, expects",
			[
					(URLPath() / pathlib.PurePosixPath("news"), URLPath("news")),
					(URLPath() / URLPath("news"), URLPath("news")),
					(URLPath('/') / pathlib.PurePosixPath("news"), URLPath("/news")),
					(URLPath('/') / URLPath("news"), URLPath("/news")),
					(URLPath('/') / MyPathLike(), URLPath("/python")),
					(URLPath("/programmes") / pathlib.PurePosixPath("b006qtlx"), URLPath("/programmes/b006qtlx")),
					(URLPath("/programmes") / URLPath("b006qtlx"), URLPath("/programmes/b006qtlx")),
					pytest.param(
							pathlib.PurePosixPath("/programmes") / URLPath("b006qtlx"),
							URLPath("/programmes/b006qtlx"),
							marks=pytest.mark.xfail(reason="The type is taken from the left object.")
							),
					]
			)
	def test_division_pathlike(self, value, expects):
		assert value == expects
		assert isinstance(value, URLPath)

	@pytest.mark.parametrize(
			"value, expects",
			[
					(URLPath().joinpath("news"), URLPath("news")),
					(URLPath('/').joinpath("news"), URLPath("/news")),
					(URLPath("/programmes").joinpath("b006qtlx"), URLPath("/programmes/b006qtlx")),
					(
							URLPath("/programmes").joinpath("b006qtlx", "details"),
							URLPath("/programmes/b006qtlx/details")
							),
					]
			)
	def test_joinpath(self, value, expects):
		assert value == expects
		assert isinstance(value, URLPath)

	# pylint: disable=expression-not-assigned

	@count(100)
	def test_division_errors_number(self, count: int):
		if sys.version_info < (3, 8):
			with pytest.raises(TypeError, match=r"expected str, bytes or os.PathLike object, not int"):
				URLPath() / count
			with pytest.raises(TypeError, match=r"expected str, bytes or os.PathLike object, not float"):
				URLPath() / float(count)
		else:
			with pytest.raises(TypeError, match=r"unsupported operand type\(s\) for /: 'URLPath' and 'int'"):
				URLPath() / count  # type: ignore[operator]
			with pytest.raises(TypeError, match=r"unsupported operand type\(s\) for /: 'URLPath' and 'float'"):
				URLPath() / float(count)  # type: ignore[operator]

	@pytest.mark.parametrize("obj", [[], (), {}, set(), pytest.raises, ABC])
	def test_division_errors(self, obj):
		if sys.version_info < (3, 8):
			with pytest.raises(TypeError, match=r"expected str, bytes or os.PathLike object, not .*"):
				URLPath() / obj
		else:
			with pytest.raises(TypeError, match=r"unsupported operand type\(s\) for /: 'URLPath' and .*"):
				URLPath() / obj

	@pytest.mark.parametrize("obj", [1234, 12.34, [], (), {}, set(), pytest.raises, ABC])
	def test_rtruediv_typerror(self, obj):
		if sys.version_info < (3, 8):
			with pytest.raises(TypeError, match=r"expected str, bytes or os.PathLike object, not .*"):
				obj / URLPath()
		else:
			with pytest.raises(TypeError, match=r"unsupported operand type\(s\) for /: .* and 'URLPath'"):
				obj / URLPath()

	# pylint: enable=expression-not-assigned

	@pytest.mark.parametrize(
			"base, other",
			[
					(URLPath("/news/sport"), "/news"),
					(URLPath("/news/sport"), URLPath("/news")),
					(URLPath("/news/sport"), pathlib.PurePosixPath("/news")),
					(URLPath("news/sport"), "news"),
					(URLPath("news/sport"), URLPath("news")),
					(URLPath("news/sport"), pathlib.PurePosixPath("news")),
					]
			)
	def test_relative_to(self, base: URLPath, other):
		assert base.relative_to(other) == URLPath("sport")
		assert isinstance(base.relative_to(other), URLPath)


class _TestURL(ABC):

	@property
	@abstractmethod
	def _class(self) -> Type:
		pass

	@pytest.mark.parametrize(
			"url, scheme, netloc, path",
			[
					(
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes/player",
							"https",
							"www.bbc.co.uk",
							URLPath("/programmes/b006qtlx/episodes/player"),
							),
					(
							"www.bbc.co.uk/programmes/b006qtlx/episodes/player",
							'',
							"www.bbc.co.uk",
							URLPath("/programmes/b006qtlx/episodes/player"),
							),
					(
							"/programmes/b006qtlx/episodes/player",
							'',
							'',
							URLPath("/programmes/b006qtlx/episodes/player"),
							),
					(
							"//www.bbc.co.uk/programmes/b006qtlx/episodes/player",
							'',
							"www.bbc.co.uk",
							URLPath("/programmes/b006qtlx/episodes/player"),
							),
					]
			)
	def test_creation(self, url, scheme, netloc, path):
		url = self._class(url)
		assert url.scheme == scheme
		assert url.netloc == netloc
		assert url.path == path

	@pytest.mark.parametrize(
			"url",
			[
					"https://www.bbc.co.uk/programmes/b006qtlx/episodes/player",
					"www.bbc.co.uk/programmes/b006qtlx/episodes/player",
					"www.bbc.co.uk",
					"/programmes/b006qtlx/episodes/player",
					"programmes/b006qtlx/episodes/player",
					"https://127.0.0.1/programmes/b006qtlx/episodes/player",
					"ftp://127.0.0.1/programmes/b006qtlx/episodes/player",
					]
			)
	def test_str(self, url):
		assert str(self._class(url)) == url

	def test_division(self):
		value = self._class("https://www.bbc.co.uk/programmes/b006qtlx/episodes") / "player"
		assert value == self._class("https://www.bbc.co.uk/programmes/b006qtlx/episodes/player")
		assert isinstance(value, self._class)

		value = self._class("/programmes/b006qtlx/episodes") / "player"
		assert value == self._class("/programmes/b006qtlx/episodes/player")
		assert isinstance(value, self._class)

		value = self._class("/programmes/b006qtlx/episodes") / "/news"
		assert value == self._class("/news")
		assert isinstance(value, self._class)

		value = self._class("www.bbc.co.uk/news")
		assert self._class("www.bbc.co.uk") / "news" == value
		assert isinstance(value, self._class)

		value = self._class("/news")
		assert self._class() / "news" == value
		assert isinstance(value, self._class)

	def test_division_pathlike(self):
		value = self._class("https://www.bbc.co.uk/programmes/b006qtlx/episodes") / pathlib.PurePosixPath("player")
		assert value == self._class("https://www.bbc.co.uk/programmes/b006qtlx/episodes/player")
		assert isinstance(value, self._class)

		value = self._class("https://www.bbc.co.uk/programmes/b006qtlx/episodes") / URLPath("player")
		assert value == self._class("https://www.bbc.co.uk/programmes/b006qtlx/episodes/player")
		assert isinstance(value, self._class)

		value = self._class("https://www.bbc.co.uk/programmes/b006qtlx/episodes") / URL("player")
		assert value == self._class("https://www.bbc.co.uk/programmes/b006qtlx/episodes/player")
		assert isinstance(value, self._class)

		value = self._class("https://www.bbc.co.uk/programmes/b006qtlx/episodes") / URL("/news")
		assert value == self._class("https://www.bbc.co.uk/news")
		assert isinstance(value, self._class)

		value = self._class("https://www.bbc.co.uk/programmes/b006qtlx/episodes") / MyPathLike()
		assert value == self._class("https://www.bbc.co.uk/python")
		assert isinstance(value, self._class)

		value = self._class("/programmes/b006qtlx/episodes") / pathlib.PurePosixPath("player")
		assert value == self._class("/programmes/b006qtlx/episodes/player")
		assert isinstance(value, self._class)

		value = self._class("/programmes/b006qtlx/episodes") / URLPath("player")
		assert value == self._class("/programmes/b006qtlx/episodes/player")
		assert isinstance(value, self._class)

		value = self._class("www.bbc.co.uk/news")
		assert self._class("www.bbc.co.uk") / pathlib.PurePosixPath("news") == value
		assert isinstance(value, self._class)

		value = self._class("www.bbc.co.uk/news")
		assert self._class("www.bbc.co.uk") / URLPath("news") == value
		assert isinstance(value, self._class)

		value = self._class("/news")
		assert self._class() / pathlib.PurePosixPath("news") == value
		assert isinstance(value, self._class)

		value = self._class("/news")
		assert self._class() / URLPath("news") == value
		assert isinstance(value, self._class)

	@count(100)
	def test_division_number(self, count: int):
		assert (self._class() / count).parts[-1] == str(count)
		assert isinstance(self._class() / count, self._class)

		with pytest.raises(TypeError, match=r"unsupported operand type\(s\) for /: '.*' and 'float'"):
			self._class() / float(count)  # pylint: disable=expression-not-assigned

	@pytest.mark.parametrize("obj", [[], (), {}, set(), pytest.raises, ABC])
	def test_division_errors(self, obj):
		with pytest.raises(TypeError, match=r"unsupported operand type\(s\) for /: '.*' and .*"):
			self._class() / obj  # pylint: disable=expression-not-assigned

	@pytest.mark.parametrize("obj", [1234, 12.34, "abcdefg", [], (), {}, set(), pytest.raises, ABC])
	def test_rtruediv_typerror(self, obj):
		with pytest.raises(TypeError, match=r"unsupported operand type\(s\) for /: .* and '.*'"):
			obj / self._class()  # pylint: disable=expression-not-assigned

	def test_joinurl(self):
		value = self._class("https://www.bbc.co.uk/programmes/b006qtlx/episodes").joinurl("player")
		assert value == self._class("https://www.bbc.co.uk/programmes/b006qtlx/episodes/player")

		value = self._class("/programmes/b006qtlx/episodes").joinurl("player")
		assert value == self._class("/programmes/b006qtlx/episodes/player")

		assert self._class("www.bbc.co.uk").joinurl("news") == self._class("www.bbc.co.uk/news")
		assert self._class().joinurl("news") == self._class("/news")

		assert self._class("www.bbc.co.uk").joinurl("news", "sport") == self._class("www.bbc.co.uk/news/sport")

		expected = self._class("www.bbc.co.uk/news/sport?que=ry")
		assert self._class("www.bbc.co.uk").joinurl("news#anchor", "sport?que=ry") == expected

		expected = self._class("www.bbc.co.uk/news/sport#anchor")
		assert self._class("www.bbc.co.uk").joinurl("news?que=ry", "sport#anchor") == expected

	def test_empty_url_operations(self):
		assert self._class().name == ''
		assert self._class().suffix == ''
		assert self._class().fqdn == ''
		assert self._class().stem == ''
		assert self._class().suffixes == []
		assert self._class().port is None

	@pytest.mark.parametrize(
			"url, name",
			[
					("https://www.bbc.co.uk/programmes/b006qtlx/episodes", "episodes"),
					("www.bbc.co.uk", ''),
					("/programmes/b006qtlx/episodes", "episodes"),
					]
			)
	def test_name(self, url, name):
		assert self._class(url).name == name

	@pytest.mark.parametrize(
			"url, suffix",
			[
					("https://www.bbc.co.uk/programmes/b006qtlx/episodes", ''),
					("www.bbc.co.uk", ''),
					("/programmes/b006qtlx/episodes", ''),
					("https://imgs.xkcd.com/comics/workflow.png", ".png"),
					]
			)
	def test_suffix(self, url, suffix):
		assert self._class(url).suffix == suffix

	@pytest.mark.parametrize(
			"url, suffixes",
			[
					("https://www.bbc.co.uk/programmes/b006qtlx/episodes", []),
					("www.bbc.co.uk", []),
					("/programmes/b006qtlx/episodes", []),
					("https://imgs.xkcd.com/comics/workflow.png", [".png"]),
					(
							"https://github.com/domdfcoding/domdf_python_tools/releases/download/"
							"v0.4.8/domdf_python_tools-0.4.8.tar.gz", [".4", ".8", ".tar", ".gz"]
							),
					]
			)
	def test_suffixes(self, url, suffixes):
		assert self._class(url).suffixes == suffixes

	@pytest.mark.parametrize(
			"url, stem",
			[
					("https://www.bbc.co.uk/programmes/b006qtlx/episodes", "episodes"),
					("www.bbc.co.uk", ''),
					("/programmes/b006qtlx/episodes", "episodes"),
					("https://imgs.xkcd.com/comics/workflow.png", "workflow"),
					(
							"https://github.com/domdfcoding/domdf_python_tools/releases/download/v0.4.8/domdf_python_tools-0.4.8.tar.gz",
							"domdf_python_tools-0.4.8.tar"
							),
					]
			)
	def test_stem(self, url, stem):
		assert self._class(url).stem == stem

	def test_with_name_errors(self):
		with pytest.raises(ValueError, match=r"URLPath\(''\) has an empty name"):
			self._class("www.bbc.co.uk").with_name("bar")

		with pytest.raises(ValueError, match=r"URLPath\(''\) has an empty name"):
			self._class().with_name("bar")

	@pytest.mark.parametrize(
			"url, expected, name",
			[(
					"https://www.bbc.co.uk/programmes/b006qtlx/episodes",
					"https://www.bbc.co.uk/programmes/b006qtlx/foo",
					"foo"
					),
				(
						"https://www.bbc.co.uk/programmes/b006qtlx/episodes?que=ry",
						"https://www.bbc.co.uk/programmes/b006qtlx/foo",
						"foo"
						),
				(
						"https://www.bbc.co.uk/programmes/b006qtlx/episodes#fragment",
						"https://www.bbc.co.uk/programmes/b006qtlx/foo",
						"foo"
						),
				(
						"https://www.bbc.co.uk/programmes/b006qtlx/episodes?que=ry#fragment",
						"https://www.bbc.co.uk/programmes/b006qtlx/foo",
						"foo"
						), ("/programmes/b006qtlx/episodes", "/programmes/b006qtlx/foo", "foo"),
				("https://imgs.xkcd.com/comics/workflow.png", "https://imgs.xkcd.com/comics/baz", "baz")]
			)
	def test_with_name(self, url: str, expected: str, name: str):
		assert self._class(url).with_name(name) == self._class(expected)

	@pytest.mark.parametrize(
			"url, expected",
			[
					(
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes",
							"https://www.bbc.co.uk/programmes/b006qtlx/foo",
							),
					(
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes?que=ry",
							"https://www.bbc.co.uk/programmes/b006qtlx/foo?que=ry",
							),
					(
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes#fragment",
							"https://www.bbc.co.uk/programmes/b006qtlx/foo#fragment",
							),
					(
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes?que=ry#fragment",
							"https://www.bbc.co.uk/programmes/b006qtlx/foo?que=ry#fragment",
							),
					]
			)
	def test_with_name_inherit(self, url: str, expected: str):
		assert self._class(url).with_name("foo", inherit=True) == self._class(expected)

	@pytest.mark.parametrize(
			"url, expected, suffix",
			[(
					"https://www.bbc.co.uk/programmes/b006qtlx/episodes",
					"https://www.bbc.co.uk/programmes/b006qtlx/episodes.foo",
					".foo"
					),
				(
						"https://www.bbc.co.uk/programmes/b006qtlx/episodes?que=ry",
						"https://www.bbc.co.uk/programmes/b006qtlx/episodes.foo",
						".foo"
						),
				(
						"https://www.bbc.co.uk/programmes/b006qtlx/episodes#fragment",
						"https://www.bbc.co.uk/programmes/b006qtlx/episodes.foo",
						".foo"
						),
				(
						"https://www.bbc.co.uk/programmes/b006qtlx/episodes?que=ry#fragment",
						"https://www.bbc.co.uk/programmes/b006qtlx/episodes.foo",
						".foo"
						), ("/programmes/b006qtlx/episodes", "/programmes/b006qtlx/episodes.foo", ".foo"),
				("https://imgs.xkcd.com/comics/workflow.png", "https://imgs.xkcd.com/comics/workflow.baz", ".baz")]
			)
	def test_with_suffix(self, url: str, expected: str, suffix: str):
		assert self._class(url).with_suffix(suffix) == self._class(expected)

	@pytest.mark.parametrize(
			"url, expected",
			[
					(
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes",
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes.foo",
							),
					(
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes?que=ry",
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes.foo?que=ry",
							),
					(
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes#fragment",
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes.foo#fragment",
							),
					(
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes?que=ry#fragment",
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes.foo?que=ry#fragment",
							),
					]
			)
	def test_with_suffix_inherit(self, url: str, expected: str):
		assert self._class(url).with_suffix(".foo", inherit=True) == self._class(expected)

	@pytest.mark.parametrize(
			"url, parent",
			[
					(
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes",
							"https://www.bbc.co.uk/programmes/b006qtlx"
							),
					("/programmes/b006qtlx/episodes", "/programmes/b006qtlx"),
					("https://imgs.xkcd.com/comics/workflow.png", "https://imgs.xkcd.com/comics"),
					]
			)
	def test_parent(self, url, parent):
		assert self._class(url).parent == self._class(parent)

	@pytest.mark.parametrize(
			"url, fqdn",
			[
					("https://www.bbc.co.uk/programmes/b006qtlx/episodes", "www.bbc.co.uk"),
					("www.bbc.co.uk", "www.bbc.co.uk"),
					("/programmes/b006qtlx/episodes", ''),
					("https://imgs.xkcd.com/comics/workflow.png", "imgs.xkcd.com"),
					]
			)
	def test_fqdn(self, url, fqdn):
		assert self._class(url).fqdn == fqdn

	@pytest.mark.parametrize(
			"url, subdomain, domain, suffix, ipv4",
			[
					("https://www.bbc.co.uk/programmes/b006qtlx/episodes", "www", "bbc", "co.uk", None),
					("/programmes/b006qtlx/episodes", '', '', '', None),
					("https://www.bbc.co.uk", "www", "bbc", "co.uk", None),
					("ftp://127.0.0.1/download.zip", '', "127.0.0.1", '', IPv4Address("127.0.0.1")),
					]
			)
	def test_domain(self, url, subdomain, domain, suffix, ipv4):
		assert isinstance(self._class(url).domain, Domain)
		assert self._class(url).domain.subdomain == subdomain
		assert self._class(url).domain.domain == domain
		assert self._class(url).domain.suffix == suffix
		assert self._class(url).domain.ipv4 == ipv4

	@pytest.mark.parametrize(
			"url, port",
			[
					("https://www.bbc.co.uk/programmes/b006qtlx/episodes", None),
					("https://www.bbc.co.uk:80/programmes/b006qtlx/episodes", 80),
					("https://www.bbc.co.uk:8080/programmes/b006qtlx/episodes", 8080),
					("https://www.bbc.co.uk:443/programmes/b006qtlx/episodes", 443),
					("www.bbc.co.uk", None),
					("www.bbc.co.uk:80", 80),
					("www.bbc.co.uk:8080", 8080),
					("www.bbc.co.uk:443", 443),
					("/programmes/b006qtlx/episodes", None),
					]
			)
	def test_port(self, url, port):
		assert self._class(url).port == port

	@pytest.mark.parametrize(
			"url, expects",
			[
					("bbc.co.uk/news", "bbc.co.uk/news"),
					("https://bbc.co.uk/news", "bbc.co.uk/news"),
					("https://www.bbc.co.uk/news", "www.bbc.co.uk/news"),
					]
			)
	def test_cast_to_pathlib(self, url, expects):
		assert pathlib.Path(self._class(url)) == pathlib.Path(expects)

	@pytest.mark.parametrize(
			"url, expects",
			[
					("bbc.co.uk/news", "bbc.co.uk/news"),
					("https://bbc.co.uk/news", "bbc.co.uk/news"),
					("https://www.bbc.co.uk/news", "www.bbc.co.uk/news"),
					]
			)
	def test_fspath(self, url, expects):
		assert self._class(url).__fspath__() == expects
		assert os.fspath(self._class(url)) == expects

	def test_hash(self):
		assert hash(self._class("bbc.co.uk")) == hash(self._class("bbc.co.uk"))
		assert hash(self._class("bbc.co.uk/news")) != hash(self._class("bbc.co.uk"))
		assert {self._class("bbc.co.uk"), hash(self._class("bbc.co.uk/news"))}

	@pytest.mark.parametrize(
			"url, parts",
			[
					(
							"https://hub.docker.com/r/tobix/pywine/dockerfile",
							("https", "hub", "docker", "com", 'r', "tobix", "pywine", "dockerfile")
							),
					]
			)
	def test_parts(self, url, parts):
		assert self._class(url).parts == parts

	def test_parents(self):
		assert self._class("https://hub.docker.com/r/tobix/pywine/dockerfile").parents == (
				self._class("https://hub.docker.com/r/tobix/pywine"),
				self._class("https://hub.docker.com/r/tobix"),
				self._class("https://hub.docker.com/r"),
				self._class("https://hub.docker.com/"),
				)

	def test_notimplemented_eq(self):
		assert URL() != 7
		assert URL() != 3.14142
		assert URL() != 'a'
		assert URL() != [7, 'a', 3.14142]
		assert URL() != (7, 'a', 3.14142)
		assert URL() != {7, 'a', 3.14142}
		assert URL() != {"int": 7, "str": 'a', "float": 3.14142}

	def test_from_url(self):
		url = URL("bbc.co.uk")
		assert self._class(url) == url

	def test_with_query(self):
		url = self._class("https://api.github.com/user/domdfcoding/repos?page=2&per_page=50")
		assert url.query == {"page": ['2'], "per_page": ["50"]}

		url = self._class("https://api.github.com/user/domdfcoding/repos?name=bob")
		assert url.query == {"name": ["bob"]}

		url = self._class("https://api.github.com/user/domdfcoding/repos")
		assert url.query == {}

		url = self._class("https://api.github.com?foo=bar")
		assert url.query == {"foo": ["bar"]}
		assert (url / "users").query == {}

	def test_with_fragment(self):
		url = self._class("https://api.github.com/user/domdfcoding/repos#footer")
		assert url.fragment == "footer"

		url = self._class("https://api.github.com/user/domdfcoding/repos")
		assert url.fragment is None

		url = self._class("https://api.github.com#footer")
		assert url.fragment == "footer"
		assert (url / "users").fragment is None

	def test_relative_to(self):
		expected = URLPath("domdfcoding")
		assert self._class("https://github.com/domdfcoding").relative_to(URL("https://github.com")) == expected
		assert self._class("https://github.com/domdfcoding").relative_to(
				RequestsURL("https://github.com")
				) == expected
		assert self._class("https://github.com/domdfcoding").relative_to(
				SlumberURL("https://github.com")
				) == expected

		expected = URLPath("football")
		assert self._class("https://www.bbc.co.uk:443/news/sport/football").relative_to("/news/sport") == expected
		assert self._class("https://www.bbc.co.uk:443/news/sport/football").relative_to(
				URLPath("/news/sport")
				) == expected

		with pytest.raises(ValueError, match="'URL.relative_to' cannot be used with relative URLPath objects"):
			self._class("https://www.bbc.co.uk:443/news/sport/football").relative_to(URLPath("news/sport"))

		with pytest.raises(ValueError, match=".* does not start with .*"):
			self._class("https://github.com/domdfcoding").relative_to(URL("https://bbc.co.uk/news"))

		with pytest.raises(ValueError, match=".* does not start with .*"):
			self._class("https://www.bbc.co.uk:443/news/sport").relative_to(URL("https://bbc.co.uk/news"))

		# Perhaps not quite what was intended
		match = r".*URL\('https://bbc.co.uk/news/sport/football(/)?'\) does not start with .*URL\('news/sport'\)"
		with pytest.raises(ValueError, match=match):
			self._class("https://bbc.co.uk/news/sport/football").relative_to("news/sport")

		the_url = self._class("https://github.com/domdfcoding")
		assert the_url.path == URLPath("/domdfcoding")
		the_url.path = URLPath("domdfcoding")
		assert the_url.path == URLPath("domdfcoding")
		assert the_url.relative_to(URL("https://github.com")) == URLPath("domdfcoding")

		# Should be case insensitive (NOT case folded) per RFC 4343
		expected = URLPath("domdfcoding")
		url = self._class("https://github.com/domdfcoding")
		assert url.relative_to(self._class("https://GitHub.COM")) == expected

	def test_ordering(self):
		# f comes before t
		football = self._class("https://bbc.co.uk:443/news/sport/football")
		tennis = self._class("https://bbc.co.uk:443/news/sport/tennis")
		assert football < tennis
		assert football <= tennis
		assert tennis > football
		assert tennis >= football
		assert sorted([tennis, football]) == [football, tennis]

		# port number is sorted before path
		tennis = self._class("https://bbc.co.uk:80/news/sport/tennis")
		assert tennis < self._class("https://bbc.co.uk:443/news/sport/football")
		assert tennis <= self._class("https://bbc.co.uk:443/news/sport/football")
		assert football > tennis
		assert football >= tennis

		# scheme is sorted before path
		tennis = self._class("http://bbc.co.uk/news/sport/tennis")
		assert tennis < self._class("https://bbc.co.uk/news/sport/football")
		assert tennis <= self._class("https://bbc.co.uk/news/sport/football")
		assert self._class("https://bbc.co.uk/news/sport/football") > tennis
		assert self._class("https://bbc.co.uk/news/sport/football") >= tennis

		# Empty subdomain comes first
		tennis = self._class("https://bbc.co.uk/news/sport/tennis")
		assert tennis < self._class("https://news.bbc.co.uk/sport/tennis")
		assert tennis <= self._class("https://news.bbc.co.uk/sport/tennis")
		assert self._class("https://news.bbc.co.uk/sport/tennis") > tennis
		assert self._class("https://news.bbc.co.uk/sport/tennis") >= tennis

	@pytest.mark.parametrize("other_class", [URL, SlumberURL, RequestsURL, TrailingRequestsURL])
	def test_ordering_other_classes(self, other_class):
		# f comes before t
		football = "https://bbc.co.uk:443/news/sport/football"
		tennis = "https://bbc.co.uk:443/news/sport/tennis"
		assert self._class(football) < other_class(tennis)
		assert self._class(football) <= other_class(tennis)
		assert self._class(tennis) > other_class(football)
		assert self._class(tennis) >= other_class(football)
		assert sorted([self._class(tennis), other_class(football)]) == [self._class(football), self._class(tennis)]

		# port number is sorted before path
		tennis = "https://bbc.co.uk:80/news/sport/tennis"
		assert self._class(tennis) < other_class(football)
		assert self._class(tennis) <= other_class(football)
		assert self._class(football) > other_class(tennis)
		assert self._class(football) >= other_class(tennis)

		# scheme is sorted before path
		tennis = "http://bbc.co.uk/news/sport/tennis"
		assert self._class(tennis) < other_class("https://bbc.co.uk/news/sport/football")
		assert self._class(tennis) <= other_class("https://bbc.co.uk/news/sport/football")
		assert self._class("https://bbc.co.uk/news/sport/football") > other_class(tennis)
		assert self._class("https://bbc.co.uk/news/sport/football") >= other_class(tennis)

		# Empty subdomain comes first
		tennis = "https://bbc.co.uk/news/sport/tennis"

		assert self._class(tennis) < other_class("https://news.bbc.co.uk/sport/tennis")
		assert self._class(tennis) <= other_class("https://news.bbc.co.uk/sport/tennis")
		assert self._class("https://news.bbc.co.uk/sport/tennis") > other_class(tennis)
		assert self._class("https://news.bbc.co.uk/sport/tennis") >= other_class(tennis)


class TestURL(_TestURL):

	_class = URL

	@pytest.mark.parametrize(
			"url, expects",
			[
					(
							URL("https://www.bbc.co.uk/programmes/b006qtlx/episodes/player"),
							"URL('https://www.bbc.co.uk/programmes/b006qtlx/episodes/player')"
							),
					(
							URL("www.bbc.co.uk/programmes/b006qtlx/episodes/player"),
							"URL('www.bbc.co.uk/programmes/b006qtlx/episodes/player')"
							),
					(URL("www.bbc.co.uk"), "URL('www.bbc.co.uk')"),
					(URL("/programmes/b006qtlx/episodes/player"), "URL('/programmes/b006qtlx/episodes/player')"),
					(URL("programmes/b006qtlx/episodes/player"), "URL('programmes/b006qtlx/episodes/player')"),
					(
							URL("https://127.0.0.1/programmes/b006qtlx/episodes/player"),
							"URL('https://127.0.0.1/programmes/b006qtlx/episodes/player')"
							),
					(
							URL("ftp://127.0.0.1/programmes/b006qtlx/episodes/player"),
							"URL('ftp://127.0.0.1/programmes/b006qtlx/episodes/player')"
							),
					]
			)
	def test_repr(self, url, expects):
		assert repr(url) == expects

	def test_isinstance(self):
		assert isinstance(URL(), URL)
		assert isinstance(URL(), os.PathLike)

	def test_equality(self):
		assert URL() == URL()
		assert URL("bbc.co.uk") == URL("bbc.co.uk")
		assert URL("https://bbc.co.uk") == URL("https://bbc.co.uk")
		assert URL("bbc.co.uk/news") == URL("bbc.co.uk/news")

		assert URL("bbc.co.uk") != URL("bbc.co.uk/news")
		assert URL("bbc.co.uk") != URL("http://bbc.co.uk/news")
		assert URL("bbc.co.uk") != URL("http://bbc.co.uk")

	def test_strict_equality(self):
		assert URL().strict_compare(URL())
		assert URL("bbc.co.uk").strict_compare(URL("bbc.co.uk"))
		assert URL("bbc.co.uk#fragment").strict_compare(URL("bbc.co.uk#fragment"))
		assert URL("bbc.co.uk?que=ry").strict_compare(URL("bbc.co.uk?que=ry"))
		assert URL("bbc.co.uk?que=ry#fragment").strict_compare(URL("bbc.co.uk?que=ry#fragment"))
		assert URL("https://bbc.co.uk").strict_compare(URL("https://bbc.co.uk"))
		assert URL("https://bbc.co.uk#fragment").strict_compare(URL("https://bbc.co.uk#fragment"))
		assert URL("https://bbc.co.uk?que=ry").strict_compare(URL("https://bbc.co.uk?que=ry"))
		assert URL("https://bbc.co.uk?que=ry#fragment").strict_compare(URL("https://bbc.co.uk?que=ry#fragment"))
		assert URL("bbc.co.uk/news").strict_compare(URL("bbc.co.uk/news"))
		assert URL("bbc.co.uk/news#fragment").strict_compare(URL("bbc.co.uk/news#fragment"))
		assert URL("bbc.co.uk/news?que=ry").strict_compare(URL("bbc.co.uk/news?que=ry"))
		assert URL("bbc.co.uk/news?que=ry#fragment").strict_compare(URL("bbc.co.uk/news?que=ry#fragment"))

		assert not URL("bbc.co.uk").strict_compare(URL("bbc.co.uk/news"))
		assert not URL("bbc.co.uk#fragment").strict_compare(URL("bbc.co.uk/news"))
		assert not URL("bbc.co.uk#fragment").strict_compare(URL("bbc.co.uk"))
		assert not URL("bbc.co.uk?que=ry").strict_compare(URL("bbc.co.uk/news"))
		assert not URL("bbc.co.uk?que=ry").strict_compare(URL("bbc.co.uk"))
		assert not URL("bbc.co.uk?que=ry").strict_compare(URL("bbc.co.uk/news#fragment"))
		assert not URL("bbc.co.uk?que=ry").strict_compare(URL("bbc.co.uk#fragment"))
		assert not URL("bbc.co.uk").strict_compare(URL("http://bbc.co.uk/news"))
		assert not URL("bbc.co.uk").strict_compare(URL("http://bbc.co.uk"))


class TestSlumberURL(_TestURL):

	_class = SlumberURL

	def test_isinstance(self):
		assert isinstance(SlumberURL(), SlumberURL)
		assert isinstance(SlumberURL(), URL)
		assert isinstance(SlumberURL(), os.PathLike)

	base = SlumberURL("https://jsonplaceholder.typicode.com")

	def test_get(self):
		assert self.base / "todos" / '1' == URL("https://jsonplaceholder.typicode.com/todos/1")
		assert (self.base / "todos" / '1').get() == {
				"userId": 1,
				"id": 1,
				"title": "delectus aut autem",
				"completed": False,
				}
		assert isinstance((self.base / "todos").get(userID=1), list)

	def test_head(self):
		assert self.base / "todos" / '1' == URL("https://jsonplaceholder.typicode.com/todos/1")
		headers = (self.base / "todos" / '1').head()

		assert "Date" in headers
		assert headers["Content-Type"] == "application/json; charset=utf-8"
		assert headers["Server"] == "cloudflare"
		# TODO: On conda brotli is used -- is the required package installed there?
		assert headers["Content-Encoding"] in {"gzip", "br"}

	def test_options(self):
		options = SlumberURL("https://readthedocs.org").options()
		assert options == "GET, HEAD, OPTIONS"

	def test_post(self):
		assert self.base / "posts" == URL("https://jsonplaceholder.typicode.com/posts")
		expected = {"body": "bar", "userId": 1, "id": 101, "title": "foo"}
		assert (self.base / "posts").post({"title": "foo", "body": "bar", "userId": 1}) == expected

	def test_put(self):
		assert self.base / "posts" / '1' == URL("https://jsonplaceholder.typicode.com/posts/1")
		expected = {"body": "bar", "userId": 1, "id": 1, "title": "foo"}
		assert (self.base / "posts" / '1').put({"title": "foo", "body": "bar", "userId": 1}) == expected

	def test_patch(self):
		body = dedent(
				"""\
		quia et suscipit
		suscipit recusandae consequuntur expedita et cum
		reprehenderit molestiae ut ut quas totam
		nostrum rerum est autem sunt rem eveniet architecto"""
				)

		expected = {"body": body, "userId": 1, "id": 1, "title": "foo"}
		assert (self.base / "posts" / '1').patch({"title": "foo"}) == expected

	def test_delete(self):
		assert (self.base / "posts" / '1').delete()

	def test_equality(self):
		assert SlumberURL() == SlumberURL()
		assert URL() == SlumberURL()

		assert SlumberURL("bbc.co.uk") == SlumberURL("bbc.co.uk")
		assert URL("bbc.co.uk") == SlumberURL("bbc.co.uk")

		assert SlumberURL("https://bbc.co.uk") == SlumberURL("https://bbc.co.uk")
		assert URL("https://bbc.co.uk") == SlumberURL("https://bbc.co.uk")

		assert SlumberURL("bbc.co.uk/news") == SlumberURL("bbc.co.uk/news")
		assert URL("bbc.co.uk/news") == SlumberURL("bbc.co.uk/news")

		assert SlumberURL("bbc.co.uk") != SlumberURL("bbc.co.uk/news")
		assert SlumberURL("bbc.co.uk") != SlumberURL("http://bbc.co.uk/news")
		assert SlumberURL("bbc.co.uk") != SlumberURL("http://bbc.co.uk")

	def test_division__store(self):
		sess = requests.Session()
		l_url = SlumberURL(
				"http://bbc.co.uk",
				format="XML",
				auth=("username", "password"),
				append_slash=False,
				session=sess,
				)
		l_url.timeout = 42
		l_url.cert = "cert"
		l_url.proxies = "proxies"  # type: ignore[assignment]
		l_url.allow_redirects = False
		l_url.verify = "verify"
		new_url = l_url / "news"
		assert new_url._store == l_url._store
		assert new_url.session is sess is l_url.session is l_url._store["session"]
		assert new_url.session.auth == l_url.session.auth == ("username", "password")
		assert new_url._store["format"] == l_url._store["format"] == "XML"
		assert new_url.timeout == 42
		assert new_url.cert == "cert"
		assert new_url.proxies == "proxies"
		assert new_url.allow_redirects is False
		assert new_url.verify == "verify"

	@not_pypy("Method never called")
	def test_garbage_collection(self, capsys):
		# Deleting a child should not close the session while a reference to it is still held

		class Sess(requests.Session):

			def close(self):
				print("Closing")
				super().close()

		u = self._class("https://github.com")
		u.session = Sess()

		u1 = u / "domdfcoding"
		print(u1 / "apeye")

		del u1
		assert capsys.readouterr().out == "https://github.com/domdfcoding/apeye\n"

		del u
		assert capsys.readouterr().out == "Closing\n"

		the_session = Sess()
		u3 = self._class("https://github.com")
		u3.session = the_session

		u4 = u3 / "domdfcoding"
		print(u4 / "apeye")

		del u4
		assert capsys.readouterr().out == "https://github.com/domdfcoding/apeye\n"

		del u3
		assert capsys.readouterr().out == ''

		the_session.close()
		assert capsys.readouterr().out == "Closing\n"


class TestRequestsURL(_TestURL):

	_class = RequestsURL

	def test_isinstance(self):
		assert isinstance(RequestsURL(), RequestsURL)
		assert isinstance(RequestsURL(), URL)
		assert isinstance(RequestsURL(), os.PathLike)

	base = RequestsURL("https://raw.githubusercontent.com")

	@pytest.mark.parametrize(
			"url, expected",
			[
					("https://github.com/domdfcoding/PyMassSpec", "https://github.com/PyMassSpec/PyMassSpec"),
					("http://pypi.io/p/domdf_python_tools", "https://pypi.org/project/domdf-python-tools"),
					]
			)
	def test_resolve(self, url: str, expected: str):
		assert str(RequestsURL(url).resolve()) == expected
		assert str(RequestsURL(url).resolve(timeout=10)) == expected

	def test_get(self):
		target_url = self.base / "domdfcoding" / "domdf_python_tools" / "v2.9.1" / "LICENSE"

		assert target_url == URL("https://raw.githubusercontent.com/domdfcoding/domdf_python_tools/v2.9.1/LICENSE")
		assert target_url.get().text.splitlines()[:11] == [
				"                   GNU LESSER GENERAL PUBLIC LICENSE",
				"                       Version 3, 29 June 2007",
				'',
				" Copyright (C) 2007 Free Software Foundation, Inc. <https://fsf.org/>",
				" Everyone is permitted to copy and distribute verbatim copies",
				" of this license document, but changing it is not allowed.",
				'',
				'',
				"  This version of the GNU Lesser General Public License incorporates",
				"the terms and conditions of version 3 of the GNU General Public",
				"License, supplemented by the additional permissions listed below.",
				]

	def test_get_default_params(self, httpserver: pytest_httpserver.pytest_plugin.PluginHTTPServer):

		class QM(QueryMatcher):

			def get_comparing_values(self, request_query_string: bytes) -> tuple:
				return '', ''

			def match(self, request_query_string: bytes) -> bool:
				return True

		def handler(request: werkzeug.Request):
			headers = {"Content-Type": "application/json"}
			body = json.dumps(parse_qs(request.query_string.decode("UTF-8")))
			return werkzeug.Response(body, 200, headers, None, None)

		httpserver.expect_request("/query", method="GET", query_string=QM()).respond_with_handler(handler)

		url = RequestsURL(httpserver.url_for("/query?foo=bar&fizz=buzz"))
		assert url.get().json() == {"foo": ["bar"], "fizz": ["buzz"]}

		url = RequestsURL(httpserver.url_for("/query"))
		assert url.get(params={"bdfl": "Guido"}).json() == {"bdfl": ["Guido"]}

	def test_requests_integration(self, httpserver: pytest_httpserver.pytest_plugin.PluginHTTPServer):
		# We only need to test the wrapper; the actual implementations are tested by requests themselves.

		httpserver.expect_request("/testing_get", method="GET").respond_with_json("GET request succeeded")
		assert RequestsURL(httpserver.url_for("/testing_get")).get().json() == "GET request succeeded"

		def setup_response(method):
			route = f"/testing_{method.lower()}"
			data = f"This is a {method} request"
			response = f"{method} request succeeded"
			expect_request = httpserver.expect_request(route, method=method, data=data)
			expect_request.respond_with_json(response)
			return httpserver.url_for(route), data, response

		route, data, response = setup_response("POST")
		assert RequestsURL(route).post(data=data).json() == response

		route, data, response = setup_response("PUT")
		assert RequestsURL(route).put(data=data).json() == response

		route, data, response = setup_response("PATCH")
		assert RequestsURL(route).patch(data=data).json() == response

		route, data, response = setup_response("OPTIONS")
		assert RequestsURL(route).options(data=data).json() == response

		method = "DELETE"
		route = f"/testing_{method.lower()}"
		data = f"This is a {method} request"
		response = f"{method} failed successfully"
		expect_request = httpserver.expect_request(route, method=method, data=data)
		expect_request.respond_with_json(response)
		assert RequestsURL(httpserver.url_for(route)).delete(data=data).json() == response

		method = "HEAD"
		route = f"/testing_{method.lower()}"
		data = f"This is a {method} request"
		test_string = "HEAD request succeeded"
		httpserver.expect_request(route, method=method, data=data).respond_with_json(test_string)
		headers = RequestsURL(httpserver.url_for(route)).head(data=data).headers
		assert headers["Content-Type"] == "application/json"
		assert headers["Content-Length"] == str(len(repr(test_string)))
		assert headers["Server"].startswith("Werkzeug")
		assert "Date" in headers

	def test_equality(self):

		assert RequestsURL() == RequestsURL()
		assert URL() == RequestsURL()

		assert RequestsURL("bbc.co.uk") == RequestsURL("bbc.co.uk")
		assert URL("bbc.co.uk") == RequestsURL("bbc.co.uk")

		assert RequestsURL("https://bbc.co.uk") == RequestsURL("https://bbc.co.uk")
		assert URL("https://bbc.co.uk") == RequestsURL("https://bbc.co.uk")

		assert RequestsURL("bbc.co.uk/news") == RequestsURL("bbc.co.uk/news")
		assert URL("bbc.co.uk/news") == RequestsURL("bbc.co.uk/news")

		assert RequestsURL("bbc.co.uk") != RequestsURL("bbc.co.uk/news")
		assert RequestsURL("bbc.co.uk") != RequestsURL("http://bbc.co.uk/news")
		assert RequestsURL("bbc.co.uk") != RequestsURL("http://bbc.co.uk")

	def test_division_session(self):
		sess = requests.Session()
		l_url = RequestsURL("http://bbc.co.uk")
		l_url.session = sess
		new_url = l_url / "news"
		assert new_url.session is sess

		sess = requests.Session()
		l_url = RequestsURL("http://bbc.co.uk")
		new_url = l_url / "news"
		assert new_url.session is not sess

	@not_pypy("Method never called")
	def test_garbage_collection(self, capsys):
		# Deleting a child should not close the session while a reference to it is still held

		class Sess(requests.Session):

			def close(self):
				print("Closing")
				super().close()

		u = RequestsURL("https://github.com")
		u.session = Sess()

		u1 = u / "domdfcoding"
		print(u1 / "apeye")

		del u1
		assert capsys.readouterr().out == "https://github.com/domdfcoding/apeye\n"

		del u
		assert capsys.readouterr().out == "Closing\n"

		the_session = Sess()
		u3 = RequestsURL("https://github.com")
		u3.session = the_session

		u4 = u3 / "domdfcoding"
		print(u4 / "apeye")

		del u4
		assert capsys.readouterr().out == "https://github.com/domdfcoding/apeye\n"

		del u3
		assert capsys.readouterr().out == ''

		the_session.close()
		assert capsys.readouterr().out == "Closing\n"


def test_subclass__eq__():

	assert RequestsURL() == SlumberURL()
	assert RequestsURL("bbc.co.uk") == SlumberURL("bbc.co.uk")
	assert RequestsURL("https://bbc.co.uk") == SlumberURL("https://bbc.co.uk")
	assert RequestsURL("bbc.co.uk/news") == SlumberURL("bbc.co.uk/news")


class TestTrailingRequestsURL(TestRequestsURL):

	_class = TrailingRequestsURL

	@pytest.mark.parametrize(
			"url, expects",
			[
					(
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes/player/",
							"https://www.bbc.co.uk/programmes/b006qtlx/episodes/player/"
							),
					(
							"www.bbc.co.uk/programmes/b006qtlx/episodes/player/",
							"www.bbc.co.uk/programmes/b006qtlx/episodes/player/"
							),
					("www.bbc.co.uk", "www.bbc.co.uk/"),
					("/programmes/b006qtlx/episodes/player", "/programmes/b006qtlx/episodes/player/"),
					("programmes/b006qtlx/episodes/player", "programmes/b006qtlx/episodes/player/"),
					(
							"https://127.0.0.1/programmes/b006qtlx/episodes/player",
							"https://127.0.0.1/programmes/b006qtlx/episodes/player/"
							),
					(
							"ftp://127.0.0.1/programmes/b006qtlx/episodes/player",
							"ftp://127.0.0.1/programmes/b006qtlx/episodes/player/"
							),
					]
			)
	def test_str(self, url, expects):
		assert str(self._class(url)) == expects

	def test_base_domain_already_trailing(self):
		assert str(TrailingRequestsURL("https://bbc.co.uk/")) == "https://bbc.co.uk/"
		assert str(TrailingRequestsURL("https://bbc.co.uk//")) == "https://bbc.co.uk/"
		assert str(TrailingRequestsURL("https://bbc.co.uk")) == "https://bbc.co.uk/"


def test_domain_class():
	d = Domain("docs", "python", "org")
	assert d.subdomain == "docs"
	assert d.domain == "python"
	assert d.suffix == "org"
	assert d.fqdn == "docs.python.org"
	assert d.registered_domain == "python.org"
	assert d.ipv4 is None
	assert repr(d) == "Domain(subdomain='docs', domain='python', suffix='org')"

	iv4d = Domain(subdomain='', domain="127.0.0.1", suffix='')
	assert iv4d.subdomain == ''
	assert iv4d.domain == "127.0.0.1"
	assert iv4d.suffix == ''
	assert iv4d.fqdn == ''
	assert iv4d.registered_domain == ''
	assert iv4d.ipv4 == IPv4Address("127.0.0.1")
	assert repr(iv4d) == "Domain(subdomain='', domain='127.0.0.1', suffix='')"



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0
#
# From https://github.com/ionrock/cachecontrol

# 3rd party
import pytest
from cachecontrol.wrapper import CacheControl
from requests import Session

# this package
from apeye.rate_limiter import RateLimitAdapter


def use_wrapper():
	print("Using helper")
	sess = CacheControl(Session())
	return sess


def use_adapter():
	print("Using adapter")
	sess = Session()
	sess.mount("http://", RateLimitAdapter())
	return sess


@pytest.fixture(params=[use_adapter, use_wrapper])
def sess(url, request):
	sess = request.param()
	sess.get(url)
	yield sess

	# closing session object
	sess.close()


class TestSessionActions:

	def test_get_caches(self, url, sess):
		r2 = sess.get(url)
		assert r2.from_cache is True

	def test_get_with_no_cache_does_not_cache(self, url, sess):
		r2 = sess.get(url, headers={"Cache-Control": "no-cache"})
		assert not r2.from_cache

	def test_put_invalidates_cache(self, url, sess):
		r2 = sess.put(url, data={"foo": "bar"})
		sess.get(url)
		assert not r2.from_cache

	def test_patch_invalidates_cache(self, url, sess):
		r2 = sess.patch(url, data={"foo": "bar"})
		sess.get(url)
		assert not r2.from_cache

	def test_delete_invalidates_cache(self, url, sess):
		r2 = sess.delete(url)
		sess.get(url)
		assert not r2.from_cache



#  Adapted from https://github.com/john-kurkowski/tldextract
#  Licensed under the BSD 3-Clause License
#
#  Copyright (c) 2020, John Kurkowski
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
from ipaddress import IPv4Address

# 3rd party
from apeye_core._tld import extract_tld

# this package
from apeye import Domain


def assert_extract(url, expected_domain_data, expected_ip_data=None):
	(expected_fqdn, expected_subdomain, expected_domain, expected_tld) = expected_domain_data
	ext = Domain._make(extract_tld(url))
	assert expected_fqdn == ext.fqdn
	assert expected_subdomain == ext.subdomain
	assert expected_domain == ext.domain
	assert expected_tld == ext.suffix
	assert expected_ip_data == ext.ipv4


def test_american():
	assert_extract("https://www.google.com", ("www.google.com", "www", "google", "com"))


def test_british():
	assert_extract("https://www.theregister.co.uk", ("www.theregister.co.uk", "www", "theregister", "co.uk"))


def test_no_subdomain():
	assert_extract("https://gmail.com", ("gmail.com", '', "gmail", "com"))


def test_nested_subdomain():
	assert_extract(
			"https://media.forums.theregister.co.uk",
			("media.forums.theregister.co.uk", "media.forums", "theregister", "co.uk")
			)


def test_odd_but_possible():
	assert_extract("https://www.www.com", ("www.www.com", "www", "www", "com"))
	assert_extract("https://www.com", ("www.com", '', "www", "com"))


def test_suffix():
	assert_extract("com", ('', '', '', "com"))
	assert_extract("co.uk", ('', '', '', "co.uk"))


def test_local_host():
	assert_extract("https://internalunlikelyhostname/", ('', '', "internalunlikelyhostname", ''))
	assert_extract("https://internalunlikelyhostname.bizarre", ('', "internalunlikelyhostname", "bizarre", ''))


def test_qualified_local_host():
	assert_extract(
			"https://internalunlikelyhostname.info/",
			("internalunlikelyhostname.info", '', "internalunlikelyhostname", "info")
			)
	assert_extract(
			"https://internalunlikelyhostname.information/", ('', "internalunlikelyhostname", "information", '')
			)


def test_ip():
	assert_extract(
			"https://216.22.0.192/",
			('', '', "216.22.0.192", ''),
			expected_ip_data=IPv4Address("216.22.0.192"),
			)
	assert_extract("https://216.22.project.coop/", ("216.22.project.coop", "216.22", "project", "coop"))


def test_looks_like_ip():
	assert_extract("1é", ('', '', "1é", ''))


def test_punycode():
	assert_extract("https://xn--h1alffa9f.xn--p1ai", ("xn--h1alffa9f.xn--p1ai", '', "xn--h1alffa9f", "xn--p1ai"))
	assert_extract("https://xN--h1alffa9f.xn--p1ai", ("xN--h1alffa9f.xn--p1ai", '', "xN--h1alffa9f", "xn--p1ai"))
	assert_extract("https://XN--h1alffa9f.xn--p1ai", ("XN--h1alffa9f.xn--p1ai", '', "XN--h1alffa9f", "xn--p1ai"))
	# Entries that might generate UnicodeError exception
	# This subdomain generates UnicodeError 'IDNA does not round-trip'
	assert_extract(
			"xn--tub-1m9d15sfkkhsifsbqygyujjrw602gk4li5qqk98aca0w.google.com",
			(
					"xn--tub-1m9d15sfkkhsifsbqygyujjrw602gk4li5qqk98aca0w.google.com",
					"xn--tub-1m9d15sfkkhsifsbqygyujjrw602gk4li5qqk98aca0w",
					"google",
					"com"
					)
			)
	# This subdomain generates UnicodeError 'incomplete punicode string'
	assert_extract(
			"xn--tub-1m9d15sfkkhsifsbqygyujjrw60.google.com",
			(
					"xn--tub-1m9d15sfkkhsifsbqygyujjrw60.google.com",
					"xn--tub-1m9d15sfkkhsifsbqygyujjrw60",
					"google",
					"com"
					)
			)


def test_invalid_puny_with_puny():
	assert_extract(
			"https://xn--zckzap6140b352by.blog.so-net.xn--wcvs22d.hk",
			(
					"xn--zckzap6140b352by.blog.so-net.xn--wcvs22d.hk",
					"xn--zckzap6140b352by.blog",
					"so-net",
					"xn--wcvs22d.hk"
					)
			)
	assert_extract("https://xn--&.so-net.com", ("xn--&.so-net.com", "xn--&", "so-net", "com"))


def test_puny_with_non_puny():
	assert_extract(
			"https://xn--zckzap6140b352by.blog.so-net.教育.hk",
			("xn--zckzap6140b352by.blog.so-net.教育.hk", "xn--zckzap6140b352by.blog", "so-net", "教育.hk")
			)


def test_idna_2008():
	"""Python supports IDNA 2003.
	The IDNA library adds 2008 support for characters like ß.
	"""
	assert_extract("xn--gieen46ers-73a.de", ("xn--gieen46ers-73a.de", '', "xn--gieen46ers-73a", "de"))


def test_empty():
	assert_extract("https://", ('', '', '', ''))


def test_scheme():
	assert_extract("https://mail.google.com/mail", ("mail.google.com", "mail", "google", "com"))
	assert_extract("ssh://mail.google.com/mail", ("mail.google.com", "mail", "google", "com"))
	assert_extract("//mail.google.com/mail", ("mail.google.com", "mail", "google", "com"))
	assert_extract("mail.google.com/mail", ("mail.google.com", "mail", "google", "com"))


def test_port():
	assert_extract("git+ssh://www.github.com:8443/", ("www.github.com", "www", "github", "com"))


def test_username():
	assert_extract("ftp://johndoe:5cr1p7k1dd13@1337.warez.com:2501", ("1337.warez.com", "1337", "warez", "com"))


def test_query_fragment():
	assert_extract("https://google.com?q=cats", ("google.com", '', "google", "com"))
	assert_extract("https://google.com#Welcome", ("google.com", '', "google", "com"))
	assert_extract("https://google.com/#Welcome", ("google.com", '', "google", "com"))
	assert_extract("https://google.com/s#Welcome", ("google.com", '', "google", "com"))
	assert_extract("https://google.com/s?q=cats#Welcome", ("google.com", '', "google", "com"))


def test_regex_order():
	assert_extract("https://www.parliament.uk", ("www.parliament.uk", "www", "parliament", "uk"))
	assert_extract("https://www.parliament.co.uk", ("www.parliament.co.uk", "www", "parliament", "co.uk"))


def test_unhandled_by_iana():
	assert_extract("https://www.cgs.act.edu.au/", ("www.cgs.act.edu.au", "www", "cgs", "act.edu.au"))
	assert_extract("https://www.google.com.au/", ("www.google.com.au", "www", "google", "com.au"))


def test_tld_is_a_website_too():
	assert_extract("https://www.metp.net.cn", ("www.metp.net.cn", "www", "metp", "net.cn"))
	# This is unhandled by the PSL. Or is it?
	# assert_extract(https://www.net.cn', ('www.net.cn', 'www', 'net', 'cn'))


def test_dns_root_label():
	assert_extract("https://www.example.com./", ("www.example.com", "www", "example", "com"))


def test_private_domains():
	assert_extract("https://waiterrant.blogspot.com", ("waiterrant.blogspot.com", "waiterrant", "blogspot", "com"))


def test_ipv4():
	assert_extract(
			"https://127.0.0.1/foo/bar", ('', '', "127.0.0.1", ''), expected_ip_data=IPv4Address("127.0.0.1")
			)


def test_ipv4_bad():
	assert_extract("https://256.256.256.256/foo/bar", ('', "256.256.256", "256", ''))


def test_ipv4_lookalike():
	assert_extract("https://127.0.0.1.9/foo/bar", ('', "127.0.0.1", '9', ''))


def test_result_as_dict():
	result = Domain._make(
			extract_tld("https://admin:password1@www.google.com:666/secret/admin/interface?param1=42")
			)
	expected_dict = {"subdomain": "www", "domain": "google", "suffix": "com"}
	assert result._asdict() == expected_dict



# stdlib
import unittest

# 3rd party
import pytest

# this package
from apeye.slumber_url import JsonSerializer, SerializerNotAvailable, SerializerRegistry, YamlSerializer


class ResourceTestCase(unittest.TestCase):

	def setUp(self):
		self.data = {"foo": "bar"}

	def test_json_get_serializer(self):
		s = SerializerRegistry()

		for content_type in [
				"application/json",
				"application/x-javascript",
				"text/javascript",
				"text/x-javascript",
				"text/x-json",
				]:
			serializer = s.get_serializer(content_type=content_type)

			if not type(serializer) == JsonSerializer:  # pylint: disable=unidiomatic-typecheck
				raise AssertionError("content_type %s should produce a JsonSerializer")

		result = serializer.dumps(self.data)
		assert result == '{"foo": "bar"}'
		assert self.data == serializer.loads(result)

	def _test_yaml_get_serializer(self):
		s = SerializerRegistry()

		for content_type in ["text/yaml"]:
			serializer = s.get_serializer(content_type=content_type)

			if not type(serializer) == YamlSerializer:  # pylint: disable=unidiomatic-typecheck
				raise AssertionError("content_type %s should produce a YamlSerializer")

		result = serializer.dumps(self.data)
		assert result == "foo: bar\n"
		assert self.data == serializer.loads(result)

	def test_yaml_get_serializer(self):
		try:
			# 3rd party
			import yaml  # noqa: F401

			self._test_yaml_get_serializer()

		except ImportError:
			try:

				# 3rd party
				import ruamel.yaml  # type: ignore[import]  # noqa: F401

				self._test_yaml_get_serializer()

			except ImportError:

				s = SerializerRegistry()

				with pytest.raises(SerializerNotAvailable, match=f"No serializer available for 'yaml'."):
					s.get_serializer("yaml")

				with pytest.raises(SerializerNotAvailable, match=f"No serializer available for 'text/yaml'."):
					s.get_serializer(content_type="text/yaml")



#  This file is managed by 'repo_helper'. Don't edit it directly.

__all__ = ["extras_require"]

extras_require = {
		"limiter": ["cachecontrol[filecache]>=0.12.6", "lockfile>=0.12.2"],
		"all": ["cachecontrol[filecache]>=0.12.6", "lockfile>=0.12.2"]
		}



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0
#
# From https://github.com/ionrock/cachecontrol

# stdlib
import os
import socket
from contextlib import suppress
from pprint import pformat

# 3rd party
import cherrypy  # type: ignore[import]
import pytest


class SimpleApp:

	def __init__(self):
		self.etag_count = 0
		self.update_etag_string()

	def dispatch(self, env):
		path = env["PATH_INFO"][1:].split('/')
		segment = path.pop(0)
		if segment and hasattr(self, segment):
			return getattr(self, segment)

		return None

	def update_etag_string(self):
		self.etag_count += 1
		self.etag_string = f'"ETAG-{self.etag_count}"'

	def __call__(self, env, start_response):
		func = self.dispatch(env)

		if func:
			return func(env, start_response)

		headers = [("Cache-Control", "max-age=5000"), ("Content-Type", "text/plain")]
		start_response("200 OK", headers)
		return [pformat(env).encode("utf8")]


@pytest.fixture(scope="session")
def server():
	return cherrypy.server


@pytest.fixture()
def url(server):
	return "http://%s:%s/" % server.bind_addr


def get_free_port():
	s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	s.bind(('', 0))
	ip, port = s.getsockname()
	s.close()
	ip = os.environ.get("WEBTEST_SERVER_BIND", "127.0.0.1")
	return ip, port


def pytest_configure(config):
	cherrypy.tree.graft(SimpleApp(), '/')

	ip, port = get_free_port()

	cherrypy.config.update({"server.socket_host": ip, "server.socket_port": port})

	# turn off logging
	logger = cherrypy.log.access_log
	logger.removeHandler(logger.handlers[0])

	cherrypy.server.start()


def pytest_unconfigure(config):
	with suppress(Exception):
		cherrypy.server.stop()
